@@ -6,13 +6,15 @@ import sys
 import pprint
 import textwrap
 import locale
-import builtins
+import glob
+import threading
 import warnings
 import contextlib
 import collections.abc as cabc
 import subprocess
 import platform
 import typing as tp
+from collections import ChainMap
 
 from xonsh import __version__ as XONSH_VERSION
 from xonsh.lazyasd import lazyobject, LazyBool
@@ -28,7 +30,7 @@ from xonsh.platform import (
     ON_CYGWIN,
     os_environ,
 )
-
+from xonsh.built_ins import XSH
 from xonsh.tools import (
     always_true,
     always_false,
@@ -60,11 +62,8 @@ from xonsh.tools import (
     csv_to_set,
     set_to_csv,
     is_int,
-    is_bool_seq,
     to_bool_or_int,
     bool_or_int_to_str,
-    csv_to_bool_seq,
-    bool_seq_to_csv,
     DefaultNotGiven,
     print_exception,
     intensify_colors_on_win_setter,
@@ -88,6 +87,7 @@ from xonsh.tools import (
     dict_to_str,
     to_int_or_none,
     DefaultNotGivenType,
+    to_repr_pretty_,
 )
 from xonsh.ansi_colors import (
     ansi_color_escape_code_to_name,
@@ -96,6 +96,7 @@ from xonsh.ansi_colors import (
     ansi_style_by_name,
 )
 import xonsh.prompt.base as prompt
+from xonsh.prompt.gitstatus import _DEFS as GITSTATUS_FIELD_DEFS
 
 events.doc(
     "on_envvar_new",
@@ -191,12 +192,8 @@ def to_debug(x):
     execer's debug level.
     """
     val = to_bool_or_int(x)
-    if (
-        hasattr(builtins, "__xonsh__")
-        and hasattr(builtins.__xonsh__, "execer")
-        and builtins.__xonsh__.execer is not None
-    ):
-        builtins.__xonsh__.execer.debug_level = val
+    if XSH.execer is not None:
+        XSH.execer.debug_level = val
     return val
 
 
@@ -395,14 +392,7 @@ class LsColors(cabc.MutableMapping):
     def __repr__(self):
         return "{0}.{1}(...)".format(self.__class__.__module__, self.__class__.__name__)
 
-    def _repr_pretty_(self, p, cycle):
-        name = "{0}.{1}".format(self.__class__.__module__, self.__class__.__name__)
-        with p.group(0, name + "(", ")"):
-            if cycle:
-                p.text("...")
-            elif len(self):
-                p.break_()
-                p.pretty(dict(self))
+    _repr_pretty_ = to_repr_pretty_
 
     def is_target(self, key) -> bool:
         """Return True if key is 'target'"""
@@ -432,7 +422,7 @@ class LsColors(cabc.MutableMapping):
     @property
     def style_name(self):
         """Current XONSH_COLOR_STYLE value"""
-        env = getattr(builtins.__xonsh__, "env", {})
+        env = getattr(XSH, "env", {}) or {}
         env_style_name = env.get("XONSH_COLOR_STYLE", "default")
         if self._style_name is None or self._style_name != env_style_name:
             self._style_name = env_style_name
@@ -484,8 +474,8 @@ class LsColors(cabc.MutableMapping):
         if filename is not None:
             cmd.append(filename)
         # get env
-        if hasattr(builtins, "__xonsh__") and hasattr(builtins.__xonsh__, "env"):
-            denv = builtins.__xonsh__.env.detype()
+        if XSH.env:
+            denv = XSH.env.detype()
         else:
             denv = None
         # run dircolors
@@ -495,6 +485,8 @@ class LsColors(cabc.MutableMapping):
             )
         except (subprocess.CalledProcessError, FileNotFoundError):
             return cls(cls.default_settings)
+        if not out:
+            return cls(cls.default_settings)
         s = out.splitlines()[0]
         _, _, s = s.partition("'")
         s, _, _ = s.rpartition("'")
@@ -524,7 +516,7 @@ def ensure_ls_colors_in_env(spec=None, **kwargs):
     environment. This fires exactly once upon the first time the
     ls command is called.
     """
-    env = builtins.__xonsh__.env
+    env = XSH.env
     if "LS_COLORS" not in env._d:
         # this adds it to the env too
         default_lscolors(env)
@@ -610,6 +602,15 @@ def default_xonshrc(env):
     return dxrc
 
 
+@default_value
+def default_xonshrcdir(env):
+    xdgrcd = os.path.join(xonsh_config_dir(env), "rc.d")
+    if ON_WINDOWS:
+        return (os.path.join(os_environ["ALLUSERSPROFILE"], "xonsh", "rc.d"), xdgrcd)
+    else:
+        return ("/etc/xonsh/rc.d", xdgrcd)
+
+
 @default_value
 def xonsh_append_newline(env):
     """Appends a newline if we are in interactive mode"""
@@ -785,6 +786,10 @@ class GeneralSetting(Xettings):
         6000,
         "Number of files on the PATH above which a warning is shown.",
     )
+    COMMANDS_CACHE_SAVE_INTERMEDIATE = Var.with_default(
+        False,
+        "If enabled, the CommandsCache saved between runs and can reduce the startup time.",
+    )
 
     HOSTNAME = Var.with_default(
         default=default_value(lambda env: platform.node()),
@@ -809,17 +814,6 @@ class GeneralSetting(Xettings):
     if hasattr(locale, "LC_MESSAGES"):
         LC_MESSAGES = Var.for_locale("LC_MESSAGES")
 
-    LOADED_RC_FILES = Var(
-        is_bool_seq,
-        csv_to_bool_seq,
-        bool_seq_to_csv,
-        (),
-        "Whether or not any of the xonsh run control files were loaded at "
-        "startup. This is a sequence of bools in Python that is converted "
-        "to a CSV list in string form, ie ``[True, False]`` becomes "
-        "``'True,False'``.",
-        is_configurable=False,
-    )
     OLDPWD = Var.with_default(
         ".",
         "Used to represent a previous present working directory.",
@@ -857,6 +851,10 @@ class GeneralSetting(Xettings):
         "should cause an end to execution. This is less useful at a terminal. "
         "The error that is raised is a ``subprocess.CalledProcessError``.",
     )
+    XONSH_SUBPROC_CAPTURED_PRINT_STDERR = Var.with_default(
+        False,
+        "If ``True`` the stderr from captured subproc will be printed automatically.",
+    )
     TERM = Var.no_default(
         "str",
         "TERM is sometimes set by the terminal emulator. This is used (when "
@@ -875,28 +873,33 @@ class GeneralSetting(Xettings):
         "not always happen.",
         is_configurable=False,
     )
+    XONSH_CAPTURE_ALWAYS = Var.with_default(
+        False,
+        "Try to capture output of commands run without explicit capturing.\n"
+        "If True, xonsh will capture the output of commands run directly or in ``![]``.\n"
+        "Setting to True has the following disadvantages:\n"
+        "* Some interactive commands won't work properly (like when ``git`` invokes an interactive editor).\n"
+        "  For more information see discussion at https://github.com/xonsh/xonsh/issues/3672.\n"
+        "* Stopping these commands with ^Z (i.e. ``SIGTSTP``)\n"
+        "  is disabled as it causes deadlocked terminals.\n"
+        "  ``SIGTSTP`` may still be issued and only the physical pressing\n"
+        "  of ``Ctrl+Z`` is ignored.\n\n"
+        "Regardless of this value, commands run in ``$()``, ``!()`` or with an IO redirection (``>`` or ``|``) "
+        "will always be captured.\n"
+        "Setting this to True depends on ``$THREAD_SUBPROCS`` being True.",
+    )
     THREAD_SUBPROCS = Var(
         is_bool_or_none,
         to_bool_or_none,
         bool_or_none_to_str,
         not ON_CYGWIN,
+        "Note: The ``$XONSH_CAPTURE_ALWAYS`` variable introduces finer control "
+        "and you should probably use that instead.\n\n"
         "Whether or not to try to run subrocess mode in a Python thread, "
-        "when applicable. There are various trade-offs, which normally "
-        "affects only interactive sessions.\n\nWhen True:\n\n"
-        "* Xonsh is able capture & store the stdin, stdout, and stderr \n"
+        "when trying to capture its output. There are various trade-offs.\n\n"
+        "If True, xonsh is able capture & store the stdin, stdout, and stderr \n"
         "  of threadable subprocesses.\n"
-        "* However, stopping threaded subprocs with ^Z (i.e. ``SIGTSTP``)\n"
-        "  is disabled as it causes deadlocked terminals.\n"
-        "  ``SIGTSTP`` may still be issued and only the physical pressing\n"
-        "  of ``Ctrl+Z`` is ignored.\n"
-        "* Threadable commands are run with ``PopenThread`` and threadable \n"
-        "  aliases are run with ``ProcProxyThread``.\n\n"
-        "When False:\n\n"
-        "* Xonsh may not be able to capture stdin, stdout, and stderr streams \n"
-        "  unless explicitly asked to do so.\n"
-        "* Stopping the thread with ``Ctrl+Z`` yields to job control.\n"
-        "* Threadable commands are run with ``Popen`` and threadable \n"
-        "  alias are run with ``ProcProxy``.\n\n"
+        "The disadvantages are listed in ``$XONSH_CAPTURE_ALWAYS``.\n"
         "The desired effect is often up to the command, user, or use case.\n\n"
         "None values are for internal use only and are used to turn off "
         "threading when loading xonshrc files. This is done because Bash "
@@ -938,6 +941,18 @@ class GeneralSetting(Xettings):
         ),
         type_str="env_path",
     )
+    XONSHRC_DIR = Var.with_default(
+        default_xonshrcdir,
+        "A list of directories, from which all .xsh files will be loaded "
+        "at startup, sorted in lexographic order. Files in these directories "
+        "are loaded after any files in XONSHRC.",
+        doc_default=(
+            "On Linux & Mac OSX: ``['/etc/xonsh/rc.d', '~/.config/xonsh/rc.d']``\n"
+            "On Windows: "
+            "``['%ALLUSERSPROFILE%\\\\xonsh\\\\rc.d', '~/.config/xonsh/rc.d']``"
+        ),
+        type_str="env_path",
+    )
     XONSH_APPEND_NEWLINE = Var.with_default(
         xonsh_append_newline,
         "Append new line when a partial line is preserved in output.",
@@ -979,13 +994,16 @@ class GeneralSetting(Xettings):
         bool_or_int_to_str,
         0,
         "Sets the xonsh debugging level. This may be an integer or a boolean. "
-        "Setting this variable prior to stating xonsh to ``1`` or ``True`` "
-        "will suppress amalgamated imports. Setting it to ``2`` will get some "
-        "basic information like input transformation, command replacement. "
-        "With ``3`` or a higher number will make more debugging information "
+        "Setting it to ``1`` will get some basic information like input transformation, command replacement. "
+        "With ``2`` or a higher number will make more debugging information "
         "presented, like PLY parsing messages.",
         is_configurable=False,
     )
+    XONSH_NO_AMALGAMATE = Var.with_default(
+        False,
+        "Setting this variable prior to starting xonsh to a truthy value will suppress amalgamated imports.",
+        is_configurable=False,
+    )
     XONSH_DATA_DIR = Var.with_default(
         xonsh_data_dir,
         "This is the location where xonsh data files are stored, such as " "history.",
@@ -1042,11 +1060,6 @@ class GeneralSetting(Xettings):
         "Whether or not to store the stdin that is supplied to the "
         "``!()`` and ``![]`` operators.",
     )
-    XONSH_STORE_STDOUT = Var.with_default(
-        False,
-        "Whether or not to store the ``stdout`` and ``stderr`` streams in the "
-        "history files.",
-    )
     XONSH_STYLE_OVERRIDES = Var(
         is_str_str_dict,
         to_str_str_dict,
@@ -1063,6 +1076,23 @@ class GeneralSetting(Xettings):
         False,
         "Set to ``True`` to show arguments list of every executed subprocess command.",
     )
+    XONSH_TRACE_COMPLETIONS = Var.with_default(
+        False,
+        "Set to ``True`` to show completers invoked and their return values.",
+    )
+    XONSH_TRACE_SUBPROC_FUNC = Var.with_default(
+        None,
+        doc=(
+            "A callback function used to format the trace output shown when $XONSH_TRACE_SUBPROC=True."
+        ),
+        doc_default="""\
+By default it just prints ``cmds`` like below.
+
+.. code-block:: python
+def tracer(cmds: list, captured: Union[bool, str]):
+    print(f"TRACE SUBPROC: {cmds}, captured={captured}", file=sys.stderr)
+""",
+    )
     XONSH_TRACEBACK_LOGFILE = Var(
         is_logfile_opt,
         to_logfile_opt,
@@ -1107,6 +1137,14 @@ class ChangeDirSetting(Xettings):
         False,
         "Whether or not to suppress directory stack manipulation output.",
     )
+    COMPLETE_DOTS = Var.with_default(
+        "matching",
+        doc="Flag to specify how current and previous directories should be "
+        "tab completed  ('./', '../'):"
+        "    - ``always`` Always complete paths with ./ and ../\n"
+        "    - ``never`` Never complete paths with ./ and ../\n"
+        "    - ``matching`` Complete if path starts with . or ..",
+    )
 
 
 class InterpreterSetting(Xettings):
@@ -1173,21 +1211,6 @@ class PromptSetting(Xettings):
         "The string used to show a shortened directory in a shortened cwd, "
         "e.g. ``'…'``.",
     )
-    HISTCONTROL = Var(
-        is_string_set,
-        csv_to_set,
-        set_to_csv,
-        set(),
-        "A set of strings (comma-separated list in string form) of options "
-        "that determine what commands are saved to the history list. By "
-        "default all commands are saved. The option ``ignoredups`` will not "
-        "save the command if it matches the previous command. The option "
-        "``ignoreerr`` will cause any commands that fail (i.e. return non-zero "
-        "exit status) to not be added to the history list. The option "
-        "``erasedups`` will remove all previous commands that matches and updates the frequency. "
-        "Note: ``erasedups`` is supported only in sqlite backend).",
-        can_store_as_str=True,
-    )
     IGNOREEOF = Var.with_default(
         False,
         "Prevents Ctrl-D from exiting the shell.",
@@ -1337,13 +1360,15 @@ class PromptSetting(Xettings):
         False,
         "Whether or not to suppress branch timeout warning messages when getting {gitstatus} PROMPT_FIELD.",
     )
-    TITLE = Var.with_default(
+    TITLE = Var(
+        is_string_or_callable,
+        ensure_string,
+        ensure_string,
         DEFAULT_TITLE,
         "The title text for the window in which xonsh is running. Formatted "
         "in the same manner as ``$PROMPT``, see 'Customizing the Prompt' "
         "http://xon.sh/tutorial.html#customizing-the-prompt.",
         doc_default="``xonsh.environ.DEFAULT_TITLE``",
-        type_str="str",
     )
     UPDATE_PROMPT_ON_KEYPRESS = Var.with_default(
         False,
@@ -1371,36 +1396,26 @@ class PromptSetting(Xettings):
         is_configurable=False,
     )
     XONSH_GITSTATUS_ = Var.with_default(
-        "",
+        None,
         "Symbols for gitstatus prompt. Default values are: \n\n"
-        "* ``XONSH_GITSTATUS_HASH``: ``:``\n"
-        "* ``XONSH_GITSTATUS_BRANCH``: ``{CYAN}``\n"
-        "* ``XONSH_GITSTATUS_OPERATION``: ``{CYAN}``\n"
-        "* ``XONSH_GITSTATUS_STAGED``: ``{RED}●``\n"
-        "* ``XONSH_GITSTATUS_CONFLICTS``: ``{RED}×``\n"
-        "* ``XONSH_GITSTATUS_CHANGED``: ``{BLUE}+``\n"
-        "* ``XONSH_GITSTATUS_UNTRACKED``: ``…``\n"
-        "* ``XONSH_GITSTATUS_STASHED``: ``⚑``\n"
-        "* ``XONSH_GITSTATUS_CLEAN``: ``{BOLD_GREEN}✓``\n"
-        "* ``XONSH_GITSTATUS_AHEAD``: ``↑·``\n"
-        "* ``XONSH_GITSTATUS_BEHIND``: ``↓·``\n",
+        + "\n".join(
+            (
+                f"* ``XONSH_GITSTATUS_{fld.name}``: ``{fld.value}``"
+                for fld in GITSTATUS_FIELD_DEFS
+            )
+        ),
         pattern="XONSH_GITSTATUS_*",
     )
-    XONSH_HISTORY_BACKEND = Var(
-        is_history_backend,
-        to_itself,
-        ensure_string,
-        "json",
-        "Set which history backend to use. Options are: 'json', "
-        "'sqlite', and 'dummy'. The default is 'json'. "
-        "``XONSH_HISTORY_BACKEND`` also accepts a class type that inherits "
-        "from ``xonsh.history.base.History``, or its instance.",
-    )
-    XONSH_HISTORY_FILE = Var.with_default(
-        os.path.expanduser("~/.xonsh_history.json"),
-        "Location of history file (deprecated).",
-        is_configurable=False,
-        doc_default="``~/.xonsh_history``",
+    XONSH_GITSTATUS_FIELDS_HIDDEN = Var.with_default(
+        (),
+        "Fields to hide in {gitstatus} prompt (all fields below are shown by default.) \n\n"
+        + "\n".join(
+            (
+                f"* ``{fld.name}``\n"
+                for fld in GITSTATUS_FIELD_DEFS
+                if not fld.name.startswith("HASH")
+            )
+        ),
     )
     XONSH_HISTORY_MATCH_ANYWHERE = Var.with_default(
         False,
@@ -1408,20 +1423,6 @@ class PromptSetting(Xettings):
         "match command history anywhere in a given line (not just the start)",
         doc_default="False",
     )
-    XONSH_HISTORY_SIZE = Var(
-        is_history_tuple,
-        to_history_tuple,
-        history_tuple_to_str,
-        (8128, "commands"),
-        "Value and units tuple that sets the size of history after garbage "
-        "collection. Canonical units are:\n\n"
-        "- ``commands`` for the number of past commands executed,\n"
-        "- ``files`` for the number of history files to keep,\n"
-        "- ``s`` for the number of seconds in the past that are allowed, and\n"
-        "- ``b`` for the number of bytes that history may consume.\n\n"
-        "Common abbreviations, such as '6 months' or '1 GB' are also allowed.",
-        doc_default="``(8128, 'commands')`` or ``'8128 commands'``",
-    )
     XONSH_STDERR_PREFIX = Var.with_default(
         "",
         "A format string, using the same keys and colors as ``$PROMPT``, that "
@@ -1440,6 +1441,67 @@ class PromptSetting(Xettings):
     )
 
 
+class PromptHistorySetting(Xettings):
+    """Interactive Prompt History"""
+
+    XONSH_HISTORY_BACKEND = Var(
+        is_history_backend,
+        to_itself,
+        ensure_string,
+        "json",
+        "Set which history backend to use. Options are: 'json', "
+        "'sqlite', and 'dummy'. The default is 'json'. "
+        "``XONSH_HISTORY_BACKEND`` also accepts a class type that inherits "
+        "from ``xonsh.history.base.History``, or its instance.",
+    )
+    XONSH_HISTORY_FILE = Var.with_default(
+        None,
+        "Location of history file set by history backend (default) or set by the user.",
+        is_configurable=False,
+        doc_default="None",
+        type_str="path",
+    )
+    HISTCONTROL = Var(
+        is_string_set,
+        csv_to_set,
+        set_to_csv,
+        set(),
+        "A set of strings (comma-separated list in string form) of options "
+        "that determine what commands are saved to the history list. By "
+        "default all commands are saved. The option ``ignoredups`` will not "
+        "save the command if it matches the previous command. The option "
+        "``ignoreerr`` will cause any commands that fail (i.e. return non-zero "
+        "exit status) to not be added to the history list. The option "
+        "``erasedups`` will remove all previous commands that matches and updates the frequency. "
+        "Note: ``erasedups`` is supported only in sqlite backend).",
+        can_store_as_str=True,
+    )
+    XONSH_HISTORY_SIZE = Var(
+        is_history_tuple,
+        to_history_tuple,
+        history_tuple_to_str,
+        (8128, "commands"),
+        "Value and units tuple that sets the size of history after garbage "
+        "collection. Canonical units are:\n\n"
+        "- ``commands`` for the number of past commands executed,\n"
+        "- ``files`` for the number of history files to keep,\n"
+        "- ``s`` for the number of seconds in the past that are allowed, and\n"
+        "- ``b`` for the number of bytes that history may consume.\n\n"
+        "Common abbreviations, such as '6 months' or '1 GB' are also allowed.",
+        doc_default="``(8128, 'commands')`` or ``'8128 commands'``",
+    )
+    XONSH_STORE_STDOUT = Var.with_default(
+        False,
+        "Whether or not to store the ``stdout`` and ``stderr`` streams in the "
+        "history files.",
+    )
+    XONSH_HISTORY_SAVE_CWD = Var.with_default(
+        True,
+        "Save current working directory to the history.",
+        doc_default="True",
+    )
+
+
 class PTKSetting(PromptSetting):  # sub-classing -> sub-group
     """Prompt Toolkit shell
     Only usable with ``$SHELL_TYPE=prompt_toolkit.``
@@ -1489,6 +1551,18 @@ class PTKSetting(PromptSetting):  # sub-classing -> sub-group
         "Whether Xonsh will auto-insert matching parentheses, brackets, and "
         "quotes. Only available under the prompt-toolkit shell.",
     )
+    XONSH_COPY_ON_DELETE = Var.with_default(
+        False,
+        "Whether to copy words/lines to clipboard on deletion (must be set in .xonshrc file)."
+        "Only available under the prompt-toolkit shell.",
+    )
+    XONSH_CTRL_BKSP_DELETION = Var.with_default(
+        False,
+        "Delete a word on CTRL-Backspace (like ALT-Backspace). "
+        r"This will only work when your terminal emulator sends ``\x7f`` on backspace and "
+        r"``\x08`` on CTRL-Backspace (which is configurable on most terminal emulators). "
+        r"On windows, the keys are reversed.",
+    )
 
 
 class AsyncPromptSetting(PTKSetting):
@@ -1520,6 +1594,12 @@ class AsyncPromptSetting(PTKSetting):
 class AutoCompletionSetting(Xettings):
     """Tab-completion behavior."""
 
+    ALIAS_COMPLETIONS_OPTIONS_BY_DEFAULT = Var.with_default(
+        doc="If True, Argparser based alias completions will show options (e.g. -h, ...) without "
+        "requesting explicitly with option prefix (-).",
+        default=False,
+        type_str="bool",
+    )
     BASH_COMPLETIONS = Var.with_default(
         doc="This is a list (or tuple) of strings that specifies where the "
         "``bash_completion`` script may be found. "
@@ -1696,7 +1776,7 @@ class Env(cabc.MutableMapping):
 
     def __init__(self, *args, **kwargs):
         """If no initial environment is given, os_environ is used."""
-        self._d = {}
+        self._d = InternalEnvironDict()
         # sentinel value for non existing envvars
         self._no_value = object()
         self._orig_env = None
@@ -1874,17 +1954,19 @@ class Env(cabc.MutableMapping):
         """Provides a context manager for temporarily swapping out certain
         environment variables with other values. On exit from the context
         manager, the original values are restored.
+        The changes are only applied to the current thread, so that they don't leak between threads.
+        To get the thread-local overrides use `get_swapped_values` and `set_swapped_values`.
         """
         old = {}
         # single positional argument should be a dict-like object
         if other is not None:
             for k, v in other.items():
                 old[k] = self.get(k, NotImplemented)
-                self[k] = v
+                self._set_item(k, v, thread_local=True)
         # kwargs could also have been sent in
         for k, v in kwargs.items():
             old[k] = self.get(k, NotImplemented)
-            self[k] = v
+            self._set_item(k, v, thread_local=True)
 
         exception = None
         try:
@@ -1895,12 +1977,18 @@ class Env(cabc.MutableMapping):
             # restore the values
             for k, v in old.items():
                 if v is NotImplemented:
-                    del self[k]
+                    self._del_item(k, thread_local=True)
                 else:
-                    self[k] = v
+                    self._set_item(k, v, thread_local=True)
             if exception is not None:
                 raise exception from None
 
+    def get_swapped_values(self):
+        return self._d.get_local_overrides()
+
+    def set_swapped_values(self, swapped_values):
+        self._d.set_local_overrides(swapped_values)
+
     #
     # Mutable mapping interface
     #
@@ -1924,6 +2012,9 @@ class Env(cabc.MutableMapping):
         return val
 
     def __setitem__(self, key, val):
+        self._set_item(key, val)
+
+    def _set_item(self, key, val, thread_local=False):
         validator = self.get_validator(key)
         converter = self.get_converter(key)
         detyper = self.get_detyper(key)
@@ -1931,7 +2022,10 @@ class Env(cabc.MutableMapping):
             val = converter(val)
         # existing envvars can have any value including None
         old_value = self._d[key] if key in self._d else self._no_value
-        self._d[key] = val
+        if thread_local:
+            self._d.set_locally(key, val)
+        else:
+            self._d[key] = val
         self._detyped = None
         if self.get("UPDATE_OS_ENVIRON"):
             if self._orig_env is None:
@@ -1948,8 +2042,14 @@ class Env(cabc.MutableMapping):
             events.on_envvar_change.fire(name=key, oldvalue=old_value, newvalue=val)
 
     def __delitem__(self, key):
+        self._del_item(key)
+
+    def _del_item(self, key, thread_local=False):
         if key in self._d:
-            del self._d[key]
+            if thread_local:
+                self._d.del_locally(key)
+            else:
+                del self._d[key]
             self._detyped = None
             if self.get("UPDATE_OS_ENVIRON") and key in os_environ:
                 del os_environ[key]
@@ -2000,9 +2100,12 @@ class Env(cabc.MutableMapping):
     def __repr__(self):
         return "{0}.{1}(...)".format(self.__class__.__module__, self.__class__.__name__)
 
+    def __hash__(self) -> int:
+        return hash(str(self._d))
+
     def _repr_pretty_(self, p, cycle):
-        name = "{0}.{1}".format(self.__class__.__module__, self.__class__.__name__)
-        with p.group(0, name + "(", ")"):
+        name = f"{self.__class__.__module__}.{self.__class__.__name__}"
+        with p.group(1, name + "(", ")"):
             if cycle:
                 p.text("...")
             elif len(self):
@@ -2091,6 +2194,81 @@ class Env(cabc.MutableMapping):
         self._vars.pop(name)
 
 
+class InternalEnvironDict(ChainMap):
+    """A dictionary which supports thread-local overrides.
+    There are two reasons we can't use ChainMap directly:
+    1. To use thread-local storage, we need to access the local().__dict__ directly each time to get the correct dict.
+    2. We want to set items to the local storage only if they were explicitly set there.
+    """
+
+    def __init__(self):
+        self._global = {}
+        self._thread_local = threading.local()
+        super().__init__()
+
+    @property
+    def _local(self):
+        # As per local's documentation, accessing its `__dict__` works fine (but must be done inside the thread).
+        return self._thread_local.__dict__
+
+    @property  # type: ignore
+    def maps(self):
+        # The 'maps' array needs to contain the thread-local dictionary every time we use it.
+        # We prefer getting from the local scope if possible.
+        return [self._local, self._global]
+
+    @maps.setter
+    def maps(self, _v):
+        # This is here for ChainMap.__init__.
+        pass
+
+    def __setitem__(self, key, value):
+        # If the value is overridden locally, set it locally.
+        local = self._local
+        if key in local:
+            local[key] = value
+        else:
+            self._global[key] = value
+
+    def __delitem__(self, key):
+        # If the value is overridden locally, delete it locally.
+        try:
+            del self._local[key]
+        except KeyError:
+            del self._global[key]
+
+    def pop(self, key, *args):
+        # If the value is overridden locally, pop it locally.
+        try:
+            return self._local.pop(key)
+        except KeyError:
+            return self._global.pop(key, *args)
+
+    def popitem(self):
+        # Fallback to the global dictionary if nothing is overridden locally.
+        try:
+            return self._local.popitem()
+        except KeyError:
+            return self._global.popitem()
+
+    def set_locally(self, key, value):
+        self._local[key] = value
+
+    def del_locally(self, key):
+        try:
+            del self._local[key]
+        except KeyError:
+            pass
+
+    def get_local_overrides(self):
+        return self._local.copy()
+
+    def set_local_overrides(self, new_local):
+        local = self._local
+        local.clear()
+        local.update(new_local)
+
+
 def _yield_executables(directory, name):
     if ON_WINDOWS:
         base_name, ext = os.path.splitext(name.lower())
@@ -2107,27 +2285,41 @@ def _yield_executables(directory, name):
 
 def locate_binary(name):
     """Locates an executable on the file system."""
-    return builtins.__xonsh__.commands_cache.locate_binary(name)
+    return XSH.commands_cache.locate_binary(name)
 
 
-def xonshrc_context(rcfiles=None, execer=None, ctx=None, env=None, login=True):
-    """Attempts to read in all xonshrc files and return the context."""
-    loaded = env["LOADED_RC_FILES"] = []
+def xonshrc_context(
+    rcfiles=None, rcdirs=None, execer=None, ctx=None, env=None, login=True
+):
+    """
+    Attempts to read in all xonshrc files (and search xonshrc directories),
+    and returns the list of rc file paths successfully loaded, in the order
+    of loading.
+    """
+    loaded = []
     ctx = {} if ctx is None else ctx
-    if rcfiles is None:
-        return env
     orig_thread = env.get("THREAD_SUBPROCS")
     env["THREAD_SUBPROCS"] = None
-    env["XONSHRC"] = tuple(rcfiles)
-    for rcfile in rcfiles:
-        if not os.path.isfile(rcfile):
-            loaded.append(False)
-            continue
-        status = xonsh_script_run_control(rcfile, ctx, env, execer=execer, login=login)
-        loaded.append(status)
+    if rcfiles is not None:
+        for rcfile in rcfiles:
+            if os.path.isfile(rcfile):
+                status = xonsh_script_run_control(
+                    rcfile, ctx, env, execer=execer, login=login
+                )
+                if status:
+                    loaded.append(rcfile)
+    if rcdirs is not None:
+        for rcdir in rcdirs:
+            if os.path.isdir(rcdir):
+                for rcfile in sorted(glob.glob(os.path.join(rcdir, "*.xsh"))):
+                    status = xonsh_script_run_control(
+                        rcfile, ctx, env, execer=execer, login=login
+                    )
+                    if status:
+                        loaded.append(rcfile)
     if env["THREAD_SUBPROCS"] is None:
         env["THREAD_SUBPROCS"] = orig_thread
-    return ctx
+    return loaded
 
 
 def windows_foreign_env_fixes(ctx):
