@@ -72,7 +72,8 @@ class HistoryEntry(types.SimpleNamespace):
     ts: two-tuple of floats
         The timestamps of when the command started and finished, including
         fractions.
-
+    cwd: str
+        The current working directory before execution the command.
     """
 
 
@@ -121,6 +122,7 @@ class History:
         self.rtns = None
         self.tss = None
         self.outs = None
+        self.cwds = None
         self.last_cmd_rtn = None
         self.last_cmd_out = None
         self.hist_size = None
@@ -141,15 +143,17 @@ class History:
                 out=self.outs[item],
                 rtn=self.rtns[item],
                 ts=self.tss[item],
+                cwd=self.cwds[item],
             )
         elif isinstance(item, slice):
             cmds = self.inps[item]
             outs = self.outs[item]
             rtns = self.rtns[item]
             tss = self.tss[item]
+            cwds = self.cwds[item]
             return [
-                HistoryEntry(cmd=c, out=o, rtn=r, ts=t)
-                for c, o, r, t in zip(cmds, outs, rtns, tss)
+                HistoryEntry(cmd=c, out=o, rtn=r, ts=t, cwd=cwd)
+                for c, o, r, t, cwd in zip(cmds, outs, rtns, tss, cwds)
             ]
         else:
             raise TypeError(
@@ -248,10 +252,11 @@ class DummyHistory(History):
 os = _LazyModule.load('os', 'os')
 sys = _LazyModule.load('sys', 'sys')
 time = _LazyModule.load('time', 'time')
-builtins = _LazyModule.load('builtins', 'builtins')
 # amalgamated collections
 threading = _LazyModule.load('threading', 'threading')
 cabc = _LazyModule.load('collections', 'collections.abc', 'cabc')
+from xonsh.built_ins import XSH
+
 try:
     import ujson as json
 except ImportError:
@@ -276,7 +281,7 @@ def _xhj_gc_commands_to_rmfiles(hsize, files):
     """
     n = 0
     ncmds = 0
-    for _, fcmds, f, _ in reversed(files):
+    for _, fcmds, _, _ in reversed(files):
         # `files` comes in with empty files included (now), don't need special handling to gc them here.
 
         if ncmds + fcmds > hsize:
@@ -286,7 +291,7 @@ def _xhj_gc_commands_to_rmfiles(hsize, files):
 
     cmds_removed = 0
     files_removed = files[:-n]
-    for _, fcmds, f, _ in files_removed:
+    for _, fcmds, _, _ in files_removed:
         cmds_removed += fcmds
 
     return cmds_removed, files_removed
@@ -299,54 +304,77 @@ def _xhj_gc_files_to_rmfiles(hsize, files):
 
 
 def _xhj_gc_seconds_to_rmfiles(hsize, files):
-    """Return age removed (the age of the *oldest* file) and list of history files to remove to get under the age limit."""
+    """Return excess duration and list of history files to remove to get under the age limit."""
     now = time.time()
-    oldest_ts = files[0][0]
     n = 0
 
-    for ts, _, f, _ in files:
+    for ts, _, _, _ in files:
         if (now - ts) < hsize:
             break
         n += 1
 
-    return (now - oldest_ts) if n > 0 else 0, files[:n]
+    rmfiles = files[:n]
+    size_over = now - hsize - rmfiles[0][0] if n > 0 else 0
+    return size_over, rmfiles
 
 
 def _xhj_gc_bytes_to_rmfiles(hsize, files):
     """Return the history files to remove to get under the byte limit."""
     n = 0
     nbytes = 0
-    for _, _, f, fsize in reversed(files):
+    for _, _, _, fsize in reversed(files):
         if nbytes + fsize > hsize:
             break
         nbytes += fsize
         n += 1
     bytes_removed = 0
     files_removed = files[:-n]
-    for _, _, f, fsize in files_removed:
+    for _, _, _, fsize in files_removed:
         bytes_removed += fsize
 
     return bytes_removed, files_removed
 
 
+def _xhj_get_data_dir():
+    dir = xt.expanduser_abs_path(
+        os.path.join(XSH.env.get("XONSH_DATA_DIR"), "history_json")
+    )
+    if not os.path.exists(dir):
+        os.makedirs(dir)
+    return dir
+
+
 def _xhj_get_history_files(sort=True, newest_first=False):
     """Find and return the history files. Optionally sort files by
     modify time.
     """
-    data_dir = builtins.__xonsh__.env.get("XONSH_DATA_DIR")
-    data_dir = xt.expanduser_abs_path(data_dir)
-    try:
-        files = [
-            os.path.join(data_dir, f)
-            for f in os.listdir(data_dir)
-            if f.startswith("xonsh-") and f.endswith(".json")
-        ]
-    except OSError:
-        files = []
-        if builtins.__xonsh__.env.get("XONSH_DEBUG"):
-            xt.print_exception("Could not collect xonsh history files.")
+    data_dirs = [
+        _xhj_get_data_dir(),
+        XSH.env.get("XONSH_DATA_DIR"),  # backwards compatibility, remove in the future
+    ]
+
+    files = []
+    for data_dir in data_dirs:
+        data_dir = xt.expanduser_abs_path(data_dir)
+        try:
+            files += [
+                os.path.join(data_dir, f)
+                for f in os.listdir(data_dir)
+                if f.startswith("xonsh-") and f.endswith(".json")
+            ]
+        except OSError:
+            if XSH.env.get("XONSH_DEBUG"):
+                xt.print_exception(
+                    f"Could not collect xonsh history json files from {data_dir}"
+                )
     if sort:
         files.sort(key=lambda x: os.path.getmtime(x), reverse=newest_first)
+
+    custom_history_file = XSH.env.get("XONSH_HISTORY_FILE", None)
+    if custom_history_file:
+        custom_history_file = xt.expanduser_abs_path(custom_history_file)
+        if custom_history_file not in files:
+            files.insert(0, custom_history_file)
     return files
 
 
@@ -374,7 +402,7 @@ class JsonHistoryGC(threading.Thread):
     def run(self):
         while self.wait_for_shell:
             time.sleep(0.01)
-        env = builtins.__xonsh__.env  # pylint: disable=no-member
+        env = XSH.env  # pylint: disable=no-member
         xonsh_debug = env.get("XONSH_DEBUG", 0)
         if self.size is None:
             hsize, units = env.get("XONSH_HISTORY_SIZE")
@@ -386,7 +414,7 @@ class JsonHistoryGC(threading.Thread):
             raise ValueError("Units type {0!r} not understood".format(units))
 
         size_over, rm_files = rmfiles_fn(hsize, files)
-        hist = getattr(builtins.__xonsh__, "history", None)
+        hist = getattr(XSH, "history", None)
         if hist is not None:  # remember last gc pass history size
             hist.hist_size = size_over + hsize
             hist.hist_units = units
@@ -418,7 +446,7 @@ class JsonHistoryGC(threading.Thread):
         This is sorted by the last closed time. Returns a list of
         (file_size, timestamp, number of cmds, file name) tuples.
         """
-        env = getattr(getattr(builtins, "__xonsh__", None), "env", None)
+        env = XSH.env
         if env is None:
             return []
 
@@ -459,7 +487,7 @@ class JsonHistoryGC(threading.Thread):
                         end="",
                         file=sys.stderr,
                     )
-            except (IOError, OSError, ValueError):
+            except (OSError, ValueError):
                 continue
         files.sort()  # this sorts by elements of the tuple,
         # the first of which just happens to be file mod time.
@@ -500,7 +528,7 @@ class JsonHistoryFlusher(threading.Thread):
 
     def dump(self):
         """Write the cached history to external storage."""
-        opts = builtins.__xonsh__.env.get("HISTCONTROL")
+        opts = XSH.env.get("HISTCONTROL", "")
         last_inp = None
         cmds = []
         for cmd in self.buffer:
@@ -522,9 +550,11 @@ class JsonHistoryFlusher(threading.Thread):
         load_hist_len = len(hist["cmds"])
         hist["cmds"].extend(cmds)
         if self.at_exit:
-            hist["ts"][1] = time.time()  # apply end time
+            # todo: check why this is here.
+            if "ts" in hist:
+                hist["ts"][1] = time.time()  # apply end time
             hist["locked"] = False
-        if not builtins.__xonsh__.env.get("XONSH_STORE_STDOUT", False):
+        if not XSH.env.get("XONSH_STORE_STDOUT", False):
             [cmd.pop("out") for cmd in hist["cmds"][load_hist_len:] if "out" in cmd]
         with open(self.filename, "w", newline="\n") as f:
             xlj.ljdump(hist, f, sort_keys=True)
@@ -595,7 +625,15 @@ class JsonHistory(History):
     JsonHistory implements an extra action: ``diff``
     """
 
-    def __init__(self, filename=None, sessionid=None, buffersize=100, gc=True, **meta):
+    def __init__(
+        self,
+        filename=None,
+        sessionid=None,
+        buffersize=100,
+        gc=True,
+        save_cwd=None,
+        **meta,
+    ):
         """Represents a xonsh session's history as an in-memory buffer that is
         periodically flushed to disk.
 
@@ -603,7 +641,7 @@ class JsonHistory(History):
         ----------
         filename : str, optional
             Location of history file, defaults to
-            ``$XONSH_DATA_DIR/xonsh-{sessionid}.json``.
+            ``$XONSH_DATA_DIR/history_json/xonsh-{sessionid}.json``.
         sessionid : int, uuid, str, optional
             Current session identifier, will generate a new sessionid if not
             set.
@@ -618,13 +656,24 @@ class JsonHistory(History):
         super().__init__(sessionid=sessionid, **meta)
         if filename is None:
             # pylint: disable=no-member
-            data_dir = builtins.__xonsh__.env.get("XONSH_DATA_DIR")
-            data_dir = os.path.expanduser(data_dir)
+            data_dir = _xhj_get_data_dir()
             self.filename = os.path.join(
                 data_dir, "xonsh-{0}.json".format(self.sessionid)
             )
         else:
             self.filename = filename
+
+        if self.filename and not os.path.exists(os.path.expanduser(self.filename)):
+            meta["cmds"] = []
+            meta["sessionid"] = str(self.sessionid)
+            with open(self.filename, "w", newline="\n") as f:
+                xlj.ljdump(meta, f, sort_keys=True)
+
+            try:
+                os.chmod(self.filename, 0o600)
+            except Exception:  # pylint: disable=broad-except
+                pass
+
         self.buffer = []
         self.buffersize = buffersize
         self._queue = collections.deque()
@@ -633,23 +682,18 @@ class JsonHistory(History):
         self._skipped = 0
         self.last_cmd_out = None
         self.last_cmd_rtn = None
-
-        meta["cmds"] = []
-        meta["sessionid"] = str(self.sessionid)
-        with open(self.filename, "w", newline="\n") as f:
-            xlj.ljdump(meta, f, sort_keys=True)
-
-        try:
-            os.chmod(self.filename, 0o600)
-        except Exception:  # pylint: disable=broad-except
-            pass
-
         self.gc = JsonHistoryGC() if gc else None
         # command fields that are known
         self.tss = JsonCommandField("ts", self)
         self.inps = JsonCommandField("inp", self)
         self.outs = JsonCommandField("out", self)
         self.rtns = JsonCommandField("rtn", self)
+        self.cwds = JsonCommandField("cwd", self)
+        self.save_cwd = (
+            save_cwd
+            if save_cwd is not None
+            else XSH.env.get("XONSH_HISTORY_SAVE_CWD", True)
+        )
 
     def __len__(self):
         return self._len - self._skipped
@@ -675,7 +719,7 @@ class JsonHistory(History):
         if not self.remember_history:
             return
 
-        opts = builtins.__xonsh__.env.get("HISTCONTROL")
+        opts = XSH.env.get("HISTCONTROL", "")
         skipped_by_ignore_space = "ignorespace" in opts and cmd.get("spc")
         if skipped_by_ignore_space:
             return None
@@ -683,6 +727,9 @@ class JsonHistory(History):
         self.buffer.append(cmd)
         self._len += 1  # must come before flushing
 
+        if not self.save_cwd and "cwd" in cmd:
+            del cmd["cwd"]
+
         try:
             del cmd["spc"]
         except KeyError:
@@ -753,7 +800,7 @@ class JsonHistory(History):
                 commands = json_file.load()["cmds"]
             except (json.decoder.JSONDecodeError, ValueError):
                 # file is corrupted somehow
-                if builtins.__xonsh__.env.get("XONSH_DEBUG") > 0:
+                if XSH.env.get("XONSH_DEBUG") > 0:
                     msg = "xonsh history file {0!r} is not valid JSON"
                     print(msg.format(f), file=sys.stderr)
                 continue
@@ -772,7 +819,7 @@ class JsonHistory(History):
         data["length"] = len(self)
         data["buffersize"] = self.buffersize
         data["bufferlength"] = len(self.buffer)
-        envs = builtins.__xonsh__.env
+        envs = XSH.env
         data["gc options"] = envs.get("XONSH_HISTORY_SIZE")
         data["gc_last_size"] = f"{(self.hist_size, self.hist_units)}"
         return data
@@ -792,6 +839,7 @@ class JsonHistory(History):
         self.inps = JsonCommandField("inp", self)
         self.outs = JsonCommandField("out", self)
         self.rtns = JsonCommandField("rtn", self)
+        self.cwds = JsonCommandField("cwd", self)
         self._len = 0
         self._skipped = 0
 
@@ -803,7 +851,6 @@ class JsonHistory(History):
 #
 # -*- coding: utf-8 -*-
 """Implements the xonsh history backend via sqlite3."""
-# amalgamated builtins
 # amalgamated collections
 json = _LazyModule.load('json', 'json')
 # amalgamated os
@@ -811,6 +858,7 @@ sqlite3 = _LazyModule.load('sqlite3', 'sqlite3')
 # amalgamated sys
 # amalgamated threading
 # amalgamated time
+# amalgamated from xonsh.built_ins import XSH
 # amalgamated xonsh.history.base
 # amalgamated xonsh.tools
 XH_SQLITE_CACHE = threading.local()
@@ -819,7 +867,7 @@ XH_SQLITE_CREATED_SQL_TBL = "CREATED_SQL_TABLE"
 
 
 def _xh_sqlite_get_file_name():
-    envs = builtins.__xonsh__.env
+    envs = XSH.env
     file_name = envs.get("XONSH_HISTORY_SQLITE_FILE")
     if not file_name:
         data_dir = envs.get("XONSH_DATA_DIR")
@@ -830,7 +878,7 @@ def _xh_sqlite_get_file_name():
 def _xh_sqlite_get_conn(filename=None):
     if filename is None:
         filename = _xh_sqlite_get_file_name()
-    return sqlite3.connect(filename)
+    return sqlite3.connect(str(filename))
 
 
 def _xh_sqlite_create_history_table(cursor):
@@ -852,12 +900,14 @@ def _xh_sqlite_create_history_table(cursor):
                   sessionid TEXT,
                   out TEXT,
                   info TEXT,
-                  frequency INTEGER default 1
+                  frequency INTEGER default 1,
+                  cwd TEXT
                  )
         """.format(
                 XH_SQLITE_TABLE_NAME
             )
         )
+
         # add frequency column if not exists for backward compatibility
         try:
             cursor.execute(
@@ -868,6 +918,14 @@ def _xh_sqlite_create_history_table(cursor):
         except sqlite3.OperationalError:
             pass
 
+        # add path column if not exists for backward compatibility
+        try:
+            cursor.execute(
+                "ALTER TABLE " + XH_SQLITE_TABLE_NAME + " ADD COLUMN cwd TEXT"
+            )
+        except sqlite3.OperationalError:
+            pass
+
         # add index on inp. since we query when erasedups is True
         cursor.execute(
             f"""\
@@ -915,6 +973,8 @@ def _xh_sqlite_insert_command(cursor, cmd, sessionid, store_stdout, remove_dupli
             ("sessionid", sessionid),
         ]
     )
+    if "cwd" in cmd:
+        values["cwd"] = cmd["cwd"]
     if store_stdout and "out" in cmd:
         values["out"] = cmd["out"]
     if "info" in cmd:
@@ -936,7 +996,7 @@ def _xh_sqlite_get_count(cursor, sessionid=None):
 
 
 def _xh_sqlite_get_records(cursor, sessionid=None, limit=None, newest_first=False):
-    sql = "SELECT inp, tsb, rtn, frequency FROM xonsh_history "
+    sql = "SELECT inp, tsb, rtn, frequency, cwd FROM xonsh_history "
     params = []
     if sessionid is not None:
         sql += "WHERE sessionid = ? "
@@ -1024,7 +1084,7 @@ class SqliteHistoryGC(threading.Thread):
         if self.size is not None:
             hsize, units = xt.to_history_tuple(self.size)
         else:
-            envs = builtins.__xonsh__.env
+            envs = XSH.env
             hsize, units = envs.get("XONSH_HISTORY_SIZE")
         if units != "commands":
             print(
@@ -1041,7 +1101,7 @@ class SqliteHistoryGC(threading.Thread):
 class SqliteHistory(History):
     """Xonsh history backend implemented with sqlite3."""
 
-    def __init__(self, gc=True, filename=None, **kwargs):
+    def __init__(self, gc=True, filename=None, save_cwd=None, **kwargs):
         super().__init__(**kwargs)
         if filename is None:
             filename = _xh_sqlite_get_file_name()
@@ -1052,6 +1112,12 @@ class SqliteHistory(History):
         self.rtns = []
         self.outs = []
         self.tss = []
+        self.cwds = []
+        self.save_cwd = (
+            save_cwd
+            if save_cwd is not None
+            else XSH.env.get("XONSH_HISTORY_SAVE_CWD", True)
+        )
 
         if not os.path.exists(self.filename):
             with _xh_sqlite_get_conn(filename=self.filename) as conn:
@@ -1068,14 +1134,15 @@ class SqliteHistory(History):
     def append(self, cmd):
         if not self.remember_history:
             return
-        envs = builtins.__xonsh__.env
+        envs = XSH.env
         inp = cmd["inp"].rstrip()
         self.inps.append(inp)
         self.outs.append(cmd.get("out"))
         self.rtns.append(cmd["rtn"])
         self.tss.append(cmd.get("ts", (None, None)))
+        self.cwds.append(cmd.get("cwd", None))
 
-        opts = envs.get("HISTCONTROL")
+        opts = envs.get("HISTCONTROL", "")
         if "ignoredups" in opts and inp == self._last_hist_inp:
             # Skipping dup cmd
             return
@@ -1085,6 +1152,8 @@ class SqliteHistory(History):
         if "ignorespace" in opts and cmd.get("spc"):
             # Skipping cmd starting with space
             return
+        if not self.save_cwd and "cwd" in cmd:
+            del cmd["cwd"]
 
         try:
             del cmd["spc"]
@@ -1104,10 +1173,10 @@ class SqliteHistory(History):
 
     def all_items(self, newest_first=False, session_id=None):
         """Display all history items."""
-        for inp, ts, rtn, freq in xh_sqlite_items(
+        for inp, ts, rtn, freq, cwd in xh_sqlite_items(
             filename=self.filename, newest_first=newest_first, sessionid=session_id
         ):
-            yield {"inp": inp, "ts": ts, "rtn": rtn, "frequency": freq}
+            yield {"inp": inp, "ts": ts, "rtn": rtn, "frequency": freq, "cwd": cwd}
 
     def items(self, newest_first=False):
         """Display history items of current session."""
@@ -1122,7 +1191,7 @@ class SqliteHistory(History):
             sessionid=self.sessionid, filename=self.filename
         )
         data["all items"] = xh_sqlite_get_count(filename=self.filename)
-        envs = builtins.__xonsh__.env
+        envs = XSH.env
         data["gc options"] = envs.get("XONSH_HISTORY_SIZE")
         return data
 
@@ -1139,6 +1208,7 @@ class SqliteHistory(History):
         self.rtns = []
         self.outs = []
         self.tss = []
+        self.cwds = []
 
         xh_sqlite_wipe_session(sessionid=self.sessionid, filename=self.filename)
 
@@ -1147,27 +1217,26 @@ class SqliteHistory(History):
 #
 # -*- coding: utf-8 -*-
 """Main entry points of the xonsh history."""
-argparse = _LazyModule.load('argparse', 'argparse')
-# amalgamated builtins
 datetime = _LazyModule.load('datetime', 'datetime')
-functools = _LazyModule.load('functools', 'functools')
 # amalgamated json
 # amalgamated os
 # amalgamated sys
 # amalgamated threading
+tp = _LazyModule.load('typing', 'typing', 'tp')
+# amalgamated from xonsh.built_ins import XSH
+xcli = _LazyModule.load('xonsh', 'xonsh.cli_utils', 'xcli')
 # amalgamated xonsh.history.base
 # amalgamated xonsh.history.dummy
 # amalgamated xonsh.history.json
 # amalgamated xonsh.history.sqlite
 xdh = _LazyModule.load('xonsh', 'xonsh.diff_history', 'xdh')
-xla = _LazyModule.load('xonsh', 'xonsh.lazyasd', 'xla')
 # amalgamated xonsh.tools
 HISTORY_BACKENDS = {"dummy": DummyHistory, "json": JsonHistory, "sqlite": SqliteHistory}
 
 
 def construct_history(**kwargs):
     """Construct the history backend object."""
-    env = builtins.__xonsh__.env
+    env = XSH.env
     backend = env.get("XONSH_HISTORY_BACKEND")
     if isinstance(backend, str) and backend in HISTORY_BACKENDS:
         kls_history = HISTORY_BACKENDS[backend]
@@ -1187,14 +1256,14 @@ def construct_history(**kwargs):
 def _xh_session_parser(hist=None, newest_first=False, **kwargs):
     """Returns history items of current session."""
     if hist is None:
-        hist = builtins.__xonsh__.history
+        hist = XSH.history
     return hist.items()
 
 
 def _xh_all_parser(hist=None, newest_first=False, **kwargs):
     """Returns all history items."""
     if hist is None:
-        hist = builtins.__xonsh__.history
+        hist = XSH.history
     return hist.all_items(newest_first=newest_first)
 
 
@@ -1318,284 +1387,269 @@ def _xh_get_history(
     return cmds
 
 
-def _xh_show_history(hist, ns, stdout=None, stderr=None):
-    """Show the requested portion of shell history.
-    Accepts same parameters with `_xh_get_history`.
-    """
-    try:
-        commands = _xh_get_history(
-            ns.session,
-            slices=ns.slices,
-            start_time=ns.start_time,
-            end_time=ns.end_time,
-            datetime_format=ns.datetime_format,
-        )
-    except Exception as err:
-        print("history: error: {}".format(err), file=stderr)
-        return
-    if ns.reverse:
-        commands = reversed(list(commands))
-    end = "\0" if ns.null_byte else "\n"
-    if ns.numerate and ns.timestamp:
-        for c in commands:
-            dt = datetime.datetime.fromtimestamp(c["ts"])
-            print(
-                "{}:({}) {}".format(c["ind"], xt.format_datetime(dt), c["inp"]),
-                file=stdout,
-                end=end,
-            )
-    elif ns.numerate:
-        for c in commands:
-            print("{}: {}".format(c["ind"], c["inp"]), file=stdout, end=end)
-    elif ns.timestamp:
-        for c in commands:
-            dt = datetime.datetime.fromtimestamp(c["ts"])
-            print(
-                "({}) {}".format(xt.format_datetime(dt), c["inp"]), file=stdout, end=end
-            )
-    else:
-        for c in commands:
-            print(c["inp"], file=stdout, end=end)
-
-
-@xla.lazyobject
-def _XH_HISTORY_SESSIONS():
-    return {
-        "session": _xh_session_parser,
-        "xonsh": _xh_all_parser,
-        "all": _xh_all_parser,
-        "zsh": _xh_zsh_hist_parser,
-        "bash": _xh_bash_hist_parser,
-    }
-
-
-_XH_MAIN_ACTIONS = {
-    "show",
-    "id",
-    "file",
-    "info",
-    "diff",
-    "gc",
-    "flush",
-    "off",
-    "on",
-    "clear",
+_XH_HISTORY_SESSIONS = {
+    "session": _xh_session_parser,
+    "xonsh": _xh_all_parser,
+    "all": _xh_all_parser,
+    "zsh": _xh_zsh_hist_parser,
+    "bash": _xh_bash_hist_parser,
 }
 
 
-@functools.lru_cache()
-def _xh_create_parser():
-    """Create a parser for the "history" command."""
-    p = argparse.ArgumentParser(
-        prog="history", description="try 'history <command> --help' " "for more info"
-    )
-    subp = p.add_subparsers(title="commands", dest="action")
-    # session action
-    show = subp.add_parser(
-        "show", prefix_chars="-+", help="display history of a session, default command"
-    )
-    show.add_argument(
-        "-r",
-        dest="reverse",
-        default=False,
-        action="store_true",
-        help="reverses the direction",
-    )
-    show.add_argument(
-        "-n",
-        dest="numerate",
-        default=False,
-        action="store_true",
-        help="numerate each command",
-    )
-    show.add_argument(
-        "-t",
-        dest="timestamp",
-        default=False,
-        action="store_true",
-        help="show command timestamps",
-    )
-    show.add_argument(
-        "-T", dest="end_time", default=None, help="show only commands before timestamp"
-    )
-    show.add_argument(
-        "+T", dest="start_time", default=None, help="show only commands after timestamp"
-    )
-    show.add_argument(
-        "-f",
-        dest="datetime_format",
-        default=None,
-        help="the datetime format to be used for" "filtering and printing",
-    )
-    show.add_argument(
-        "-0",
-        dest="null_byte",
-        default=False,
-        action="store_true",
-        help="separate commands by the null character for piping "
-        "history to external filters",
-    )
-    show.add_argument(
-        "session",
-        nargs="?",
-        choices=_XH_HISTORY_SESSIONS.keys(),
-        default="session",
-        metavar="session",
-        help="{} (default: current session, all is an alias for xonsh)"
-        "".format(", ".join(map(repr, _XH_HISTORY_SESSIONS.keys()))),
-    )
-    show.add_argument(
-        "slices",
-        nargs="*",
-        default=None,
-        metavar="slice",
-        help="integer or slice notation",
-    )
-    # 'id' subcommand
-    subp.add_parser("id", help="display the current session id")
-    # 'file' subcommand
-    subp.add_parser("file", help="display the current history filename")
-    # 'info' subcommand
-    info = subp.add_parser(
-        "info", help=("display information about the " "current history")
-    )
-    info.add_argument(
-        "--json",
-        dest="json",
-        default=False,
-        action="store_true",
-        help="print in JSON format",
-    )
+class HistoryAlias(xcli.ArgParserAlias):
+    """Try 'history <command> --help' for more info"""
 
-    # gc
-    gcp = subp.add_parser("gc", help="launches a new history garbage collector")
-    gcp.add_argument(
-        "--size",
-        nargs=2,
-        dest="size",
-        default=None,
-        help=(
-            "next two arguments represent the history size and "
-            'units; e.g. "--size 8128 commands"'
-        ),
-    )
-    gcp.add_argument(
-        "--force",
-        dest="force_gc",
-        default=False,
-        action="store_true",
-        help="perform garbage collection even if history much bigger than configured limit",
-    )
-    bgcp = gcp.add_mutually_exclusive_group()
-    bgcp.add_argument(
-        "--blocking",
-        dest="blocking",
-        default=True,
-        action="store_true",
-        help=("ensures that the gc blocks the main thread, " "default True"),
-    )
-    bgcp.add_argument(
-        "--non-blocking",
-        dest="blocking",
-        action="store_false",
-        help="makes the gc non-blocking, and thus return sooner",
-    )
+    def show(
+        self,
+        session: xcli.Annotated[
+            str, xcli.Arg(nargs="?", choices=tuple(_XH_HISTORY_SESSIONS))
+        ] = "session",
+        slices: xcli.Annotated[tp.List[int], xcli.Arg(nargs="*")] = None,
+        datetime_format: tp.Optional[str] = None,
+        start_time: tp.Optional[str] = None,
+        end_time: tp.Optional[str] = None,
+        location: tp.Optional[str] = None,
+        reverse=False,
+        numerate=False,
+        timestamp=False,
+        null_byte=False,
+        _stdout=None,
+        _stderr=None,
+        _unparsed=None,
+    ):
+        """Display history of a session, default command
 
-    hist = builtins.__xonsh__.history
-    if isinstance(hist, JsonHistory):
-        # add actions belong only to JsonHistory
-        diff = subp.add_parser("diff", help="diff two xonsh history files")
-        xdh.dh_create_parser(p=diff)
+        Parameters
+        ----------
+        session:
+            The history session to get. (all is an alias for xonsh)
+        slices:
+            integer or slice notation to get only portions of history.
+        datetime_format : -f
+            the datetime format to be used for filtering and printing
+        start_time: --start-time, +T
+            show only commands after timestamp
+        end_time: -T, --end-time
+            show only commands before timestamp
+        location: -l, --location
+            The history file location (bash or zsh)
+        reverse: -r, --reverse
+            Reverses the direction
+        numerate: -n, --numerate
+            Numerate each command
+        timestamp: -t, --ts, --time-stamp
+            show command timestamps
+        null_byte: -0, --nb, --null-byte
+            separate commands by the null character for piping history to external filters
+        _unparsed
+            remaining args from ``parser.parse_known_args``
+        """
+        slices = list(slices or ())
+        if _unparsed:
+            slices.extend(_unparsed)
+        try:
+            commands = _xh_get_history(
+                session,
+                slices=slices,
+                start_time=start_time,
+                end_time=end_time,
+                datetime_format=datetime_format,
+                location=location,
+            )
+        except Exception as err:
+            self.parser.error(err)
+            return
 
-    # 'flush' subcommand
-    subp.add_parser("flush", help="flush the current history to disk")
-
-    # 'off' subcommand
-    subp.add_parser("off", help="history will not be saved for this session")
-
-    # 'on' subcommand
-    subp.add_parser(
-        "on", help="history will be saved for the rest of the session (default)"
-    )
-
-    # 'clear' subcommand
-    subp.add_parser("clear", help="one-time wipe of session history")
-
-    return p
-
-
-def _xh_parse_args(args):
-    """Prepare and parse arguments for the history command.
-
-    Add default action for ``history`` and
-    default session for ``history show``.
-    """
-    parser = _xh_create_parser()
-    if not args:
-        args = ["show", "session"]
-    elif args[0] not in _XH_MAIN_ACTIONS and args[0] not in ("-h", "--help"):
-        args = ["show", "session"] + args
-    if args[0] == "show":
-        if not any(a in _XH_HISTORY_SESSIONS for a in args):
-            args.insert(1, "session")
-        ns, slices = parser.parse_known_args(args)
-        if slices:
-            if not ns.slices:
-                ns.slices = slices
-            else:
-                ns.slices.extend(slices)
-    else:
-        ns = parser.parse_args(args)
-    return ns
-
-
-def history_main(
-    args=None, stdin=None, stdout=None, stderr=None, spec=None, stack=None
-):
-    """This is the history command entry point."""
-    hist = builtins.__xonsh__.history
-    ns = _xh_parse_args(args)
-    if not ns or not ns.action:
-        return
-    if ns.action == "show":
-        _xh_show_history(hist, ns, stdout=stdout, stderr=stderr)
-    elif ns.action == "info":
-        data = hist.info()
-        if ns.json:
-            s = json.dumps(data)
-            print(s, file=stdout)
+        if reverse:
+            commands = reversed(list(commands))
+        end = "\0" if null_byte else "\n"
+        if numerate and timestamp:
+            for c in commands:
+                dt = datetime.datetime.fromtimestamp(c["ts"])
+                print(
+                    "{}:({}) {}".format(c["ind"], xt.format_datetime(dt), c["inp"]),
+                    file=_stdout,
+                    end=end,
+                )
+        elif numerate:
+            for c in commands:
+                print("{}: {}".format(c["ind"], c["inp"]), file=_stdout, end=end)
+        elif timestamp:
+            for c in commands:
+                dt = datetime.datetime.fromtimestamp(c["ts"])
+                print(
+                    "({}) {}".format(xt.format_datetime(dt), c["inp"]),
+                    file=_stdout,
+                    end=end,
+                )
         else:
-            lines = ["{0}: {1}".format(k, v) for k, v in data.items()]
-            print("\n".join(lines), file=stdout)
-    elif ns.action == "id":
+            for c in commands:
+                print(c["inp"], file=_stdout, end=end)
+
+    @staticmethod
+    def id_cmd(_stdout):
+        """Display the current session id"""
+        hist = XSH.history
         if not hist.sessionid:
             return
-        print(str(hist.sessionid), file=stdout)
-    elif ns.action == "file":
-        if not hist.filename:
-            return
-        print(str(hist.filename), file=stdout)
-    elif ns.action == "gc":
-        hist.run_gc(size=ns.size, blocking=ns.blocking, force=ns.force_gc)
-    elif ns.action == "diff":
-        if isinstance(hist, JsonHistory):
-            xdh.dh_main_action(ns)
-    elif ns.action == "flush":
+        print(str(hist.sessionid), file=_stdout)
+
+    @staticmethod
+    def flush(_stdout):
+        """Flush the current history to disk"""
+
+        hist = XSH.history
         hf = hist.flush()
         if isinstance(hf, threading.Thread):
             hf.join()
-    elif ns.action == "off":
+
+    @staticmethod
+    def off():
+        """History will not be saved for this session"""
+        hist = XSH.history
         if hist.remember_history:
             hist.clear()
             hist.remember_history = False
             print("History off", file=sys.stderr)
-    elif ns.action == "on":
+
+    @staticmethod
+    def on():
+        """History will be saved for the rest of the session (default)"""
+        hist = XSH.history
         if not hist.remember_history:
             hist.remember_history = True
             print("History on", file=sys.stderr)
-    elif ns.action == "clear":
+
+    @staticmethod
+    def clear():
+        """One-time wipe of session history"""
+        hist = XSH.history
         hist.clear()
         print("History cleared", file=sys.stderr)
-    else:
-        print("Unknown history action {}".format(ns.action), file=sys.stderr)
+
+    @staticmethod
+    def file(_stdout):
+        """Display the current history filename"""
+        hist = XSH.history
+        if not hist.filename:
+            return
+        print(str(hist.filename), file=_stdout)
+
+    @staticmethod
+    def info(
+        to_json=False,
+        _stdout=None,
+    ):
+        """Display information about the current history
+
+        Parameters
+        ----------
+        to_json: -j, --json
+            print in JSON format
+        """
+        hist = XSH.history
+
+        data = hist.info()
+        if to_json:
+            s = json.dumps(data)
+            print(s, file=_stdout)
+        else:
+            lines = ["{0}: {1}".format(k, v) for k, v in data.items()]
+            print("\n".join(lines), file=_stdout)
+
+    @staticmethod
+    def gc(
+        size: xcli.Annotated[tp.Tuple[int, str], xcli.Arg(nargs=2)] = None,
+        force=False,
+        _blocking=True,
+    ):
+        """Launches a new history garbage collector
+
+        Parameters
+        ----------
+        size : -s, --size
+            Next two arguments represent the history size and units; e.g. "--size 8128 commands"
+        force : -f, --force
+            perform garbage collection even if history much bigger than configured limit
+        """
+        hist = XSH.history
+        hist.run_gc(size=size, blocking=_blocking, force=force)
+
+    @staticmethod
+    def diff(
+        a,
+        b,
+        reopen=False,
+        verbose=False,
+        _stdout=None,
+    ):
+        """Diff two xonsh history files
+
+        Parameters
+        ----------
+        left:
+            The first file to diff
+        right:
+            The second file to diff
+        reopen: -r, --reopen
+            make lazy file loading reopen files each time
+        verbose: -v, --verbose
+            whether to print even more information
+        """
+
+        hist = XSH.history
+        if isinstance(hist, JsonHistory):
+            hd = xdh.HistoryDiffer(a, b, reopen=reopen, verbose=verbose)
+            xt.print_color(hd.format(), file=_stdout)
+
+    def build(self):
+        parser = self.create_parser(prog="history")
+        parser.add_command(self.show, prefix_chars="-+")
+        parser.add_command(self.id_cmd, prog="id")
+        parser.add_command(self.file)
+        parser.add_command(self.info)
+        parser.add_command(self.flush)
+        parser.add_command(self.off)
+        parser.add_command(self.on)
+        parser.add_command(self.clear)
+
+        gcp = parser.add_command(self.gc)
+        bgcp = gcp.add_mutually_exclusive_group()
+        bgcp.add_argument(
+            "--blocking",
+            dest="_blocking",
+            default=True,
+            action="store_true",
+            help="ensures that the gc blocks the main thread, default True",
+        )
+        bgcp.add_argument(
+            "--non-blocking",
+            dest="_blocking",
+            action="store_false",
+            help="makes the gc non-blocking, and thus return sooner",
+        )
+        if isinstance(XSH.history, JsonHistory):
+            # add actions belong only to JsonHistory
+            parser.add_command(self.diff)
+
+        return parser
+
+    def __call__(self, args, *rest, **kwargs):
+        if not args:
+            args = ["show", "session"]
+        else:
+            actions = self.parser.commands.choices
+            cmd = args[0]
+            if cmd not in actions and cmd not in {"-h", "--help"}:
+                args = ["show", "session"] + args
+
+        if args[0] == "show":
+            if not any(a in _XH_HISTORY_SESSIONS for a in args):
+                args.insert(1, "session")
+
+            kwargs.setdefault("lenient", True)
+        return super().__call__(args, *rest, **kwargs)
+
+
+history_main = HistoryAlias()
 
