@@ -1,17 +1,18 @@
 # -*- coding: utf-8 -*-
 """Base prompt, provides PROMPT_FIELDS and prompt related functions"""
 
-import builtins
 import itertools
 import os
 import re
 import socket
 import sys
+import typing as tp
 
 import xonsh.lazyasd as xl
 import xonsh.tools as xt
 import xonsh.platform as xp
 
+from xonsh.built_ins import XSH
 from xonsh.prompt.cwd import (
     _collapsed_pwd,
     _replace_home_cwd,
@@ -29,6 +30,54 @@ def DEFAULT_PROMPT():
     return default_prompt()
 
 
+class _ParsedToken(tp.NamedTuple):
+    """It can either be a literal value alone or a field and its resultant value"""
+
+    value: str
+    field: tp.Optional[str] = None
+
+
+class ParsedTokens(tp.NamedTuple):
+    tokens: tp.List[_ParsedToken]
+    template: tp.Union[str, tp.Callable]
+
+    def process(self) -> str:
+        """Wrapper that gets formatter-function from environment and returns final prompt."""
+        processor = XSH.env.get(  # type: ignore
+            "PROMPT_TOKENS_FORMATTER", prompt_tokens_formatter_default
+        )
+        return processor(self)
+
+    def update(
+        self,
+        idx: int,
+        val: tp.Optional[str],
+        spec: tp.Optional[str],
+        conv: tp.Optional[str],
+    ) -> None:
+        """Update tokens list in-place"""
+        if idx < len(self.tokens):
+            tok = self.tokens[idx]
+            self.tokens[idx] = _ParsedToken(_format_value(val, spec, conv), tok.field)
+
+
+def prompt_tokens_formatter_default(container: ParsedTokens) -> str:
+    """
+        Join the tokens
+
+    Parameters
+    ----------
+    container: ParsedTokens
+        parsed tokens holder
+
+    Returns
+    -------
+    str
+        process the tokens and finally return the prompt string
+    """
+    return "".join([tok.value for tok in container.tokens])
+
+
 class PromptFormatter:
     """Class that holds all the related prompt formatting methods,
     uses the ``PROMPT_FIELDS`` envvar (no color formatting).
@@ -37,40 +86,45 @@ class PromptFormatter:
     def __init__(self):
         self.cache = {}
 
-    def __call__(self, template=DEFAULT_PROMPT, fields=None, **kwargs):
+    def __call__(self, template=DEFAULT_PROMPT, fields=None, **kwargs) -> str:
         """Formats a xonsh prompt template string."""
 
         # keep cache only during building prompt
         self.cache.clear()
 
         if fields is None:
-            self.fields = builtins.__xonsh__.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
+            self.fields = XSH.env.get("PROMPT_FIELDS", PROMPT_FIELDS)  # type: ignore
         else:
             self.fields = fields
         try:
-            prompt = self._format_prompt(template=template, **kwargs)
-        except Exception:
+            toks = self._format_prompt(template=template, **kwargs)
+            prompt = toks.process()
+        except Exception as ex:
+            # make it obvious why it has failed
+            print(
+                f"Failed to format prompt `{template}`-> {type(ex)}:{ex}",
+                file=sys.stderr,
+            )
             return _failover_template_format(template)
         return prompt
 
-    def _format_prompt(self, template=DEFAULT_PROMPT, **kwargs):
-        return "".join(self._get_tokens(template, **kwargs))
-
-    def _get_tokens(self, template, **kwargs):
-        template = template() if callable(template) else template
+    def _format_prompt(self, template=DEFAULT_PROMPT, **kwargs) -> ParsedTokens:
+        tmpl = template() if callable(template) else template
         toks = []
-        for literal, field, spec, conv in xt.FORMATTER.parse(template):
-            toks.append(literal)
+        for literal, field, spec, conv in xt.FORMATTER.parse(tmpl):
+            if literal:
+                toks.append(_ParsedToken(literal))
             entry = self._format_field(field, spec, conv, idx=len(toks), **kwargs)
             if entry is not None:
-                toks.append(entry)
-        return toks
+                toks.append(_ParsedToken(entry, field))
 
-    def _format_field(self, field, spec, conv, **kwargs):
+        return ParsedTokens(toks, template)
+
+    def _format_field(self, field, spec="", conv=None, **kwargs):
         if field is None:
             return
         elif field.startswith("$"):
-            val = builtins.__xonsh__.env[field[1:]]
+            val = XSH.env[field[1:]]
             return _format_value(val, spec, conv)
         elif field in self.fields:
             val = self._get_field_value(field, spec=spec, conv=conv, **kwargs)
@@ -172,7 +226,7 @@ def multiline_prompt(curr=""):
     # tail is the trailing whitespace
     tail = line if headlen == 0 else line.rsplit(head[-1], 1)[1]
     # now to construct the actual string
-    dots = builtins.__xonsh__.env.get("MULTILINE_PROMPT")
+    dots = XSH.env.get("MULTILINE_PROMPT")
     dots = dots() if callable(dots) else dots
     if dots is None or len(dots) == 0:
         return ""
@@ -218,7 +272,7 @@ def is_template_string(template, PROMPT_FIELDS=None):
         return False
     included_names.discard(None)
     if PROMPT_FIELDS is None:
-        fmtter = builtins.__xonsh__.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
+        fmtter = XSH.env.get("PROMPT_FIELDS", PROMPT_FIELDS)
     else:
         fmtter = PROMPT_FIELDS
     known_names = set(fmtter.keys())
